"""Controller holding and managing Vantage 'back boxes'."""

from aiovantage.controllers.base import BaseController
from aiovantage.objects import BackBox


class BackBoxesController(BaseController[BackBox]):
    """Controller holding and managing Vantage 'back boxes'.

    Back boxes typically represent a "gang box" in a wall which may contain
    multiple stations, dimmers, etc. It is mostly useful to know about these
    devices so we can set up a proper device hierarchy.
    """

    vantage_types = ("BackBox",)



"""Interface for querying and controlling variables."""

from dataclasses import dataclass

from .base import Interface


class GMemInterface(Interface):
    """Interface for querying and controlling variables."""

    @dataclass
    class Buffer:
        """Response from a GMem fetch operation."""

        rcode: int
        data: bytes
        size: int

    method_signatures = {
        "GMem.Fetch": Buffer,
    }

    async def fetch(self, vid: int) -> Buffer:
        """Fetch the contents of the variable.

        Args:
            vid: The Vantage ID of the variable.

        Returns:
            The contents of the variable.
        """
        # INVOKE <id> GMem.Fetch
        # -> R:INVOKE <id> <rcode> GMem.Fetch <buffer> <size>
        return await self.invoke(vid, "GMem.Fetch")

    async def commit(self, vid: int, buffer: bytes) -> None:
        """Set the contents of the variable.

        Args:
            vid: The Vantage ID of the variable.
            buffer: The contents to set the variable to.
        """
        # INVOKE <id> GMem.Commit <buffer> <size>
        # -> R:INVOKE <id> <rcode> GMem.Commit <buffer> <size>
        await self.invoke(vid, "GMem.Commit", buffer)



"""Interface for querying and controlling fans."""

from enum import IntEnum

from .base import Interface


class FanInterface(Interface):
    """Interface for querying and controlling fans."""

    class FanSpeed(IntEnum):
        """Fan speed."""

        Off = 0
        Low = 1
        Medium = 2
        Hight = 3
        Max = 4
        Auto = 5

    method_signatures = {
        "Fan.GetSpeed": FanSpeed,
        "Fan.GetSpeedHW": FanSpeed,
    }

    async def get_speed(self, vid: int, *, hw: bool = False) -> FanSpeed:
        """Get the speed of a fan.

        Args:
            vid: The Vantage ID of the fan.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The speed of the fan.
        """
        # INVOKE <id> Fan.GetSpeed
        # -> R:INVOKE <id> <speed> Fan.GetSpeed
        return await self.invoke(vid, "Fan.GetSpeedHW" if hw else "Fan.GetSpeed")

    async def set_speed(self, vid: int, speed: FanSpeed, *, sw: bool = False) -> None:
        """Set the speed of a fan.

        Args:
            vid: The Vantage ID of the fan.
            speed: The speed to set the fan to.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Fan.SetSpeed <speed>
        # -> R:INVOKE <id> <rcode> Fan.SetSpeed <speed>
        await self.invoke(vid, "Fan.SetSpeedSW" if sw else "Fan.SetSpeed", speed)

    async def increase_speed(self, vid: int) -> None:
        """Increase the speed of a fan.

        Args:
            vid: The Vantage ID of the fan.
        """
        # INVOKE <id> Fan.IncreaseSpeed
        # -> R:INVOKE <id> <rcode> Fan.IncreaseSpeed
        await self.invoke(vid, "Fan.IncreaseSpeed")

    async def decrease_speed(self, vid: int) -> None:
        """Decrease the speed of a fan.

        Args:
            vid: The Vantage ID of the fan.
        """
        # INVOKE <id> Fan.DecreaseSpeed
        # -> R:INVOKE <id> <rcode> Fan.DecreaseSpeed
        await self.invoke(vid, "Fan.DecreaseSpeed")



"""Temperature object."""

from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class Temperature(Sensor):
    """Temperature object."""

    parent: Parent
    out_of_range: int = 0
    in_range: int = 0
    range_high: float
    range_low: float
    hold_on_time: float
    temp: int = 0

    # Not available in 2.x firmware, not strictly in the schema
    class Setpoint(Enum):
        """Setpoint type."""

        HEAT = "Heat"
        COOL = "Cool"
        AUTO = "Auto"

    setpoint: Setpoint | None = field(default=None, metadata={"type": "Attribute"})

    # State
    value: Decimal | None = field(default=None, metadata={"type": "Ignore"})



"""Controller holding and managing Vantage controllers."""

from typing import Any

from typing_extensions import override

from aiovantage.object_interfaces import (
    IntrospectionInterface,
    ObjectInterface,
)
from aiovantage.objects import Master

from .base import BaseController


class MastersController(
    BaseController[Master], IntrospectionInterface, ObjectInterface
):
    """Controller holding and managing Vantage controllers."""

    vantage_types = ("Master",)
    """The Vantage object types that this controller will fetch."""

    interface_status_types = ("Object.GetMTime",)
    """Which object interface status messages this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a Vantage controller."""
        state: dict[str, Any] = {
            "firmware_version": await self.get_version(),
        }

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        if method != "Object.GetMTime":
            return

        state = {
            "m_time": self.parse_object_status(method, result, *args),
        }

        self.update_state(vid, state)

    async def get_version(self) -> str:
        """Get the firmware version of a Vantage controller.

        Returns:
            The firmware version of the controller.
        """
        # VERSION
        # -> R:VERSION {version}
        response = await self.command_client.command("VERSION")
        return response.args[0]



"""Blind base class."""

from dataclasses import dataclass, field
from decimal import Decimal

from .system_object import SystemObject


@dataclass(kw_only=True)
class BlindBase(SystemObject):
    """Blind base class."""

    shade_orientation: str | None = field(default=None, metadata={"type": "Attribute"})
    shade_type: str | None = field(default=None, metadata={"type": "Attribute"})

    # State
    position: Decimal | None = field(default=None, metadata={"type": "Ignore"})



"""Interface for querying and controlling loads."""

from decimal import Decimal
from enum import IntEnum

from .base import Interface


class LoadInterface(Interface):
    """Interface for querying and controlling loads."""

    class RampType(IntEnum):
        """Load ramp type."""

        Up = 5
        Down = 4
        Opposite = 3
        Stop = 2
        Fixed = 6
        Variable = 7
        Adjust = 8

    class AlertState(IntEnum):
        """Load alert state."""

        Clear = 0
        Overload = 1
        BulbChange = 2
        WrongType = 3
        DCCurrent = 4
        ShortCircuit = 5

    class DimmingConfig(IntEnum):
        """Load dimming config."""

        Manual = 0
        Forward = 1
        Reverse = 2
        Auto = 3

    method_signatures = {
        "Load.GetLevel": Decimal,
        "Load.GetLevelHW": Decimal,
        "Load.GetProfile": int,
        "Load.GetOverrideLevel": Decimal,
        "Load.GetAlertState": AlertState,
        "Load.GetDimmingConfig": DimmingConfig,
    }

    async def set_level(
        self, vid: int, level: float | Decimal, *, sw: bool = False
    ) -> None:
        """Set the level of a load.

        Args:
            vid: The Vantage ID of the load.
            level: The level to set the load to (0-100).
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Load.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Load.SetLevel <level (0-100)>
        await self.invoke(vid, "Load.SetLevelSW" if sw else "Load.SetLevel", level)

    async def get_level(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the level of a load.

        Args:
            vid: The Vantage ID of the load.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetLevel
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetLevel
        return await self.invoke(vid, "Load.GetLevelHW" if hw else "Load.GetLevel")

    async def ramp(
        self,
        vid: int,
        cmd: RampType,
        ramptime: float | Decimal,
        finallevel: float | Decimal,
    ) -> None:
        """Ramp a load to a level over a number of seconds.

        Args:
            vid: The Vantage ID of the load.
            cmd: The type of ramp to perform.
            ramptime: The number of seconds to ramp the load over.
            finallevel: The level to ramp the load to (0-100).
        """
        # INVOKE <id> Load.Ramp <cmd> <time> <level>
        # -> R:INVOKE <id> <rcode> Load.Ramp <cmd> <time> <level>
        await self.invoke(vid, "Load.Ramp", cmd, ramptime, finallevel)

    async def set_profile(self, vid: int, profile: int) -> None:
        """Set the id of the power profile used by this load.

        Args:
            vid: The Vantage ID of the load.
            profile: The power profile id to set the load to.
        """
        # INVOKE <id> Load.SetProfile <profile>
        # -> R:INVOKE <id> <rcode> Load.SetProfile <profile>
        await self.invoke(vid, "Load.SetProfile", profile)

    async def get_profile(self, vid: int) -> int:
        """Get the id of the power profile used by this load.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The power profile id used by the load.
        """
        # INVOKE <id> Load.GetProfile
        # -> R:INVOKE <id> <profile> Load.GetProfile
        return await self.invoke(vid, "Load.GetProfile")

    async def get_override_level(self, vid: int) -> Decimal:
        """Get the override level of a load.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The override level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetOverrideLevel
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetOverrideLevel
        return await self.invoke(vid, "Load.GetOverrideLevel")

    async def ramp_auto_off(
        self,
        vid: int,
        cmd: RampType,
        ramptime: float | Decimal,
        finallevel: float | Decimal,
        offcmd: RampType,
        offramptime: float | Decimal,
        offtimeout: float | Decimal,
    ) -> None:
        """Ramp a load to a level over a number of seconds, then ramp off after a timeout.

        Args:
            vid: The Vantage ID of the load.
            cmd: The type of ramp to perform.
            ramptime: The number of seconds to ramp the load over.
            finallevel: The level to ramp the load to (0-100).
            offcmd: The type of ramp to perform to turn the load off.
            offramptime: The number of seconds to ramp the load off over.
            offtimeout: The number of seconds to wait before turning the load off.
        """
        # INVOKE <id> Load.RampAutoOff <cmd> <time> <level> <offcmd> <offtime> <offlevel>
        # -> R:INVOKE <id> <rcode> Load.RampAutoOff <cmd> <time> <level> <offcmd> <offtime> <offlevel>
        await self.invoke(
            vid,
            "Load.RampAutoOff",
            cmd,
            ramptime,
            finallevel,
            offcmd,
            offramptime,
            offtimeout,
        )

    async def get_alert_state(self, vid: int) -> AlertState:
        """Get the alert state of a load.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The alert state of the load.
        """
        # INVOKE <id> Load.GetAlertState
        # -> R:INVOKE <id> <alert state> Load.GetAlertState
        return await self.invoke(vid, "Load.GetAlertState")

    async def set_alert_state(self, vid: int, alert_state: AlertState) -> None:
        """Set the cached alert state of a load.

        Args:
            vid: The Vantage ID of the load.
            alert_state: The alert state to set the load to.
        """
        # INVOKE <id> Load.SetAlertStateSW <alert state>
        # -> R:INVOKE <id> <rcode> Load.SetAlertStateSW <alert state>
        await self.invoke(vid, "Load.SetAlertStateSW", alert_state)

    async def get_dimming_config(self, vid: int) -> DimmingConfig:
        """Get the dimming configuration of a load.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The dimming configuration of the load.
        """
        # INVOKE <id> Load.GetDimmingConfig
        # -> R:INVOKE <id> <dimming config> Load.GetDimmingConfig
        return await self.invoke(vid, "Load.GetDimmingConfig")

    # Convenience functions, not part of the interface
    async def turn_on(
        self, vid: int, transition: float | None = None, level: float | None = None
    ) -> None:
        """Turn on a load with an optional transition time.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to transition to the new level, defaults to immediate.
            level: The level to set the load to (0-100), defaults to 100.
        """
        if level is None:
            level = 100

        if transition is None:
            return await self.set_level(vid, level)

        await self.ramp(vid, self.RampType.Fixed, transition, level)

    async def turn_off(self, vid: int, transition: float | None = None) -> None:
        """Turn off a load with an optional transition time.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to ramp the load down, defaults to immediate.
        """
        if transition is None:
            return await self.set_level(vid, 0)

        await self.ramp(vid, self.RampType.Fixed, transition, 0)



"""OmniSensor object."""

from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum

from .sensor import Sensor
from .types import Parent


class ConversionType(Enum):
    """OmniSensor type conversion information."""

    FIXED = "fixed"
    INT = "int"


@dataclass(kw_only=True)
class OmniSensor(Sensor):
    """OmniSensor object."""

    @dataclass
    class Get:
        @dataclass(kw_only=True)
        class Formula:
            return_type: ConversionType = field(
                default=ConversionType.FIXED,
                metadata={"type": "Attribute"},
            )
            level_type: ConversionType = field(
                default=ConversionType.INT,
                metadata={"type": "Attribute"},
            )
            formula: str

        formula: Formula
        method: str
        method_hw: str = field(metadata={"name": "MethodHW"})

    @dataclass
    class Set:
        @dataclass(kw_only=True)
        class Formula:
            return_type: ConversionType = field(
                default=ConversionType.INT,
                metadata={"type": "Attribute"},
            )
            value_type: ConversionType = field(
                default=ConversionType.FIXED,
                metadata={"type": "Attribute"},
            )
            formula: str

        formula: Formula
        method: str
        method_sw: str = field(metadata={"name": "MethodSW"})

    parent: Parent
    get: Get
    set: Set

    # State
    level: int | Decimal | None = field(default=None, metadata={"type": "Ignore"})

    @property
    def is_current_sensor(self) -> bool:
        """Return True if the sensor is a current sensor."""
        return self.model == "Current"

    @property
    def is_power_sensor(self) -> bool:
        """Return True if the sensor is a power sensor."""
        return self.model == "Power"

    @property
    def is_temperature_sensor(self) -> bool:
        """Return True if the sensor is a temperature sensor."""
        return self.model == "Temperature"



"""IIntrospection.GetInterfaces method definition."""

from dataclasses import dataclass, field


@dataclass
class GetInterfaces:
    """IIntrospection.GetInterfaces method definition."""

    interface = "IIntrospection"

    @dataclass
    class Interface:
        """Object interface definition."""

        name: str
        version: str
        iid: int = field(metadata={"name": "IID"})

    call = None

    result: list[Interface] | None = field(
        default_factory=list,
        metadata={
            "wrapper": "return",
            "name": "Interface",
            "type": "Element",
        },
    )



"""Interface for querying and controlling system objects."""

import datetime as dt
from dataclasses import dataclass

from .base import Interface


class ObjectInterface(Interface):
    """Interface for querying and controlling system objects."""

    @dataclass
    class PropertyEx:
        """Extended property information."""

        rcode: int
        property: str
        value: str
        size: int

    method_signatures = {
        "Object.GetVID": int,
        "Object.GetController": int,
        "Object.GetType": str,
        "Object.GetName": str,
        "Object.GetModel": str,
        "Object.GetNote": str,
        "Object.GetProperty": int,
        "Object.GetPropertyEx": PropertyEx,
        "Object.IsLocked": bool,
        "Object.IsInterfaceSupported": bool,
        "Object.IsMethodSupported": bool,
        "Object.GetMTime": dt.datetime,
        "Object.GetDName": str,
        "Object.GetArea": int,
    }

    async def get_vid(self, vid: int) -> int:
        """Get the Vantage ID of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The Vantage ID of the object.
        """
        # INVOKE <id> Object.GetVID
        # -> R:INVOKE <id> <vid> Object.GetVID
        return await self.invoke(vid, "Object.GetVID")

    async def get_controller(self, vid: int) -> int:
        """Get the VID of the controller of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The VID of the controller of the object.
        """
        # INVOKE <id> Object.GetController
        # -> R:INVOKE <id> <controller> Object.GetController
        return await self.invoke(vid, "Object.GetController")

    async def get_type(self, vid: int) -> str:
        """Get the type of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The type of the object.
        """
        # INVOKE <id> Object.GetType
        # -> R:INVOKE <id> <type> Object.GetType
        return await self.invoke(vid, "Object.GetType")

    async def get_name(self, vid: int) -> str:
        """Get the name field of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The name of the object.
        """
        # INVOKE <id> Object.GetName
        # -> R:INVOKE <id> <name> Object.GetName
        return await self.invoke(vid, "Object.GetName")

    async def get_model(self, vid: int) -> str:
        """Get the model field of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The model of the object.
        """
        # INVOKE <id> Object.GetModel
        # -> R:INVOKE <id> <model> Object.GetModel
        return await self.invoke(vid, "Object.GetModel")

    async def get_note(self, vid: int) -> str:
        """Get the note field of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The note of the object.
        """
        # INVOKE <id> Object.GetNote
        # -> R:INVOKE <id> <note> Object.GetNote
        return await self.invoke(vid, "Object.GetNote")

    async def get_property(self, vid: int, xpath: str) -> int:
        """Get an integer property of an object.

        Args:
            vid: The Vantage ID of the object.
            xpath: XPath of the property to get, eg: "DName", "Get/Formula/@ReturnType, etc.

        Returns:
            The value of the property.
        """
        # INVOKE <id> Object.GetProperty <xpath>
        # -> R:INVOKE <id> <value> Object.GetProperty <xpath>
        return await self.invoke(vid, "Object.GetProperty", xpath)

    async def get_property_ex(self, vid: int, xpath: str) -> PropertyEx:
        """Get a string property of an object.

        Args:
            vid: The Vantage ID of the object.
            xpath: XPath of the property to get, eg: "DName", "Get/Formula/@ReturnType, etc.

        Returns:
            The value of the property.
        """
        # INVOKE <id> Object.GetPropertyEx <xpath>
        # -> R:INVOKE <id> <value> Object.GetPropertyEx <xpath>
        return await self.invoke(vid, "Object.GetPropertyEx", xpath)

    async def lock(self, vid: int) -> None:
        """Lock an object.

        Args:
            vid: The Vantage ID of the object.
        """
        # INVOKE <id> Object.Lock
        # -> R:INVOKE <id> <rcode> Object.Lock
        await self.invoke(vid, "Object.Lock")

    async def unlock(self, vid: int) -> None:
        """Unlock an object.

        Args:
            vid: The Vantage ID of the object.
        """
        # INVOKE <id> Object.Unlock
        # -> R:INVOKE <id> <rcode> Object.Unlock
        await self.invoke(vid, "Object.Unlock")

    async def is_locked(self, vid: int) -> bool:
        """Check if an object is locked.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            True if the object is locked, False otherwise.
        """
        # INVOKE <id> Object.IsLocked
        # -> R:INVOKE <id> <locked (0/1)> Object.IsLocked
        return await self.invoke(vid, "Object.IsLocked")

    async def is_interface_supported(self, vid: int, iid: int) -> bool:
        """Check if an interface is supported by an object.

        Args:
            vid: The Vantage ID of the object.
            iid: The interface ID to check.

        Returns:
            True if the interface is supported, False otherwise.
        """
        # INVOKE <id> Object.IsInterfaceSupported <iid>
        # -> R:INVOKE <id> <supported (0/1)> Object.IsInterfaceSupported <iid>
        return await self.invoke(vid, "Object.IsInterfaceSupported", iid)

    async def is_method_supported(self, vid: int, iid: int, mid: int) -> bool:
        """Check if a method is supported by an object.

        Args:
            vid: The Vantage ID of the object.
            iid: The interface ID to check.
            mid: The method ID to check.

        Returns:
            True if the method is supported, False otherwise.
        """
        # INVOKE <id> Object.IsMethodSupported <iid> <mid>
        # -> R:INVOKE <id> <supported (0/1)> Object.IsMethodSupported <iid> <mid>
        return await self.invoke(vid, "Object.IsMethodSupported", iid, mid)

    async def set_property(self, vid: int, property: str, value: int) -> None:
        """Set an integer property of an object.

        Args:
            vid: The Vantage ID of the object.
            property: The property to set.
            value: The value to set the property to.
        """
        # INVOKE <id> Object.SetProperty <property> <value>
        # -> R:INVOKE <id> <rcode> Object.SetProperty <property> <value>
        await self.invoke(vid, "Object.SetProperty", property, value)

    async def set_property_ex(self, vid: int, property: str, value: str) -> None:
        """Set a string property of an object.

        Args:
            vid: The Vantage ID of the object.
            property: The property to set.
            value: The value to set the property to.
        """
        # INVOKE <id> Object.SetPropertyEx <property> <value>
        # -> R:INVOKE <id> <rcode> Object.SetPropertyEx <property> <value>
        await self.invoke(vid, "Object.SetPropertyEx", property, value)

    async def get_mtime(self, vid: int) -> dt.datetime:
        """Get the modification time of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The modification time of the object, as a datetime object.
        """
        # INVOKE <id> Object.GetMTime
        # -> R:INVOKE <id> <mtime> Object.GetMTime
        return await self.invoke(vid, "Object.GetMTime")

    async def get_dname(self, vid: int) -> str:
        """Get the display name of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The display name of the object.
        """
        # INVOKE <id> Object.GetDName
        # -> R:INVOKE <id> <dname> Object.GetDName
        return await self.invoke(vid, "Object.GetDName")

    async def get_area(self, vid: int) -> int:
        """Get the area of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The area of the object.
        """
        # INVOKE <id> Object.GetArea
        # -> R:INVOKE <id> <area> Object.GetArea
        return await self.invoke(vid, "Object.GetArea")



"""Command client data conversion utilities."""

import datetime as dt
import re
import struct
from abc import ABC, abstractmethod
from decimal import Decimal
from enum import IntEnum
from typing import Any

TOKEN_PATTERN = re.compile(r'"([^""]*(?:""[^""]*)*)"|(\{.*?\})|(\[.*?\])|(\S+)')


def tokenize_response(string: str) -> list[str]:
    """Tokenize a response from the Host Command service.

    Handles quoted strings and byte arrays as single tokens.

    Args:
        string: The response string to tokenize.

    Returns:
        A list of string tokens.
    """
    tokens: list[str] = []
    for match in TOKEN_PATTERN.finditer(string):
        token = match.group(0)

        # Remove quotes from quoted strings, and unescape quotes
        if token.startswith('"') and token.endswith('"'):
            token = token[1:-1].replace('""', '"')

        tokens.append(token)

    return tokens


class Converter(ABC):
    """Abstract data converter class."""

    @abstractmethod
    def serialize(self, value: Any, **kwargs: Any) -> str: ...  # noqa: D102

    @abstractmethod
    def deserialize(self, value: str, **kwargs: Any) -> Any: ...  # noqa: D102


class ConverterRegistry:
    """A registry for data converters."""

    registry: dict[type, Converter]

    def __init__(self):
        """Initialize the registry."""
        self.registry = {}

    def register(self, data_type: type, converter: Converter) -> None:
        """Register a converter for a data type.

        Args:
            data_type: The type of data to convert.
            converter: The converter instance.
        """
        self.registry[data_type] = converter

    def get_converter(self, data_type: type) -> Converter:
        """Get the converter for a data type.

        Args:
            data_type: The type of data to convert.

        Returns:
            The converter for the data type.
        """
        # Check if the data type is directly registered
        if data_type in self.registry:
            return self.registry[data_type]

        # If not, check the MRO for a registered type
        for mro in data_type.__mro__[1:-1]:
            if mro in self.registry:
                return self.registry[mro]

        raise ValueError(f"No converter found for {data_type}")

    def deserialize(self, data_type: type, value: str, **kwargs: Any) -> Any:
        """Deserialize a string representation to an object.

        Args:
            data_type: The type of data to convert.
            value: The string data to deserialize.
            **kwargs: Additional deserialization arguments.

        Returns:
            The deserialized object.
        """
        converter = self.get_converter(data_type)
        return converter.deserialize(value, data_type=data_type, **kwargs)

    def serialize(self, value: Any, **kwargs: Any) -> str:
        """Serialize an object to a string representation.

        Args:
            value: The value to serialize to a string.
            **kwargs: Additional serialization arguments.

        Returns:
            A string representation of the object.
        """
        converter = self.get_converter(value.__class__)
        return converter.serialize(value, **kwargs)


class StringConverter(Converter):
    """String parameter converter.

    Vantage "string" parameters are either single words or quoted strings.

    Quoted strings are wrapped in double quotes, and any double quotes within
    the string are escaped with an additional double quote.
    """

    def deserialize(self, value: str, **_kwargs: Any) -> str:
        """Deserialize a string value."""
        if value.startswith('"') and value.endswith('"'):
            return value[1:-1].replace('""', '"')

        return value

    def serialize(self, value: str, **kwargs: Any) -> str:
        """Serialize a string value."""
        force_quotes = kwargs.get("force_quotes", False)

        if '"' in value or " " in value or force_quotes:
            value = value.replace('"', '""')
            return f'"{value}"'

        return value


class BoolConverter(Converter):
    """Bool parameter converter.

    Vantage "bool" parameters are encoded as either "0" or "1".
    """

    def deserialize(self, value: str, **_kwargs: Any) -> bool:
        """Deserialize a bool value."""
        return bool(int(value))

    def serialize(self, value: bool, **_kwargs: Any) -> str:
        """Serialize a bool value."""
        return "1" if value else "0"


class IntConverter(Converter):
    """An int converter."""

    def deserialize(self, value: str, **kwargs: Any) -> int:
        """Deserialize an int value."""
        return int(value)

    def serialize(self, value: int, **kwargs: Any) -> str:
        """Serialize an int value."""
        return str(value)


class FloatConverter(Converter):
    """A float converter.

    If these types are tagged as "fixed", they should be handled as Decimal values.
    """

    def deserialize(self, value: str, **kwargs: Any) -> float:
        """Deserialize a float value."""
        return float(value)

    def serialize(self, value: float, **kwargs: Any) -> str:
        """Serialize a float value."""
        precision = kwargs.get("precision", 3)
        return f"{value:.{precision}f}"


class DecimalConverter(Converter):
    """A Decimal converter.

    Vantage "fixed" parameters are encoded as a string representation of a
    fixed-point value, with three decimal places.
    """

    def deserialize(self, value: str, **kwargs: Any) -> Decimal:
        """Deserialize a Decimal value."""
        # Handle both forms of fixed-point values:
        # - "123.456" (INVOKE replies)
        # - "123456"  (S:STATUS messages)
        return Decimal(value.replace(".", "")) / 1000

    def serialize(self, value: Decimal, **kwargs: Any) -> str:
        """Serialize a Decimal value."""
        precision = kwargs.get("precision", 3)
        return f"{value:.{precision}f}"


class BytesConverter(Converter):
    """A bytes converter.

    Vantage "bytes" parameters are encoded as a string of signed 32-bit integers,
    separated by commas or spaces, and wrapped in curly or square braces.
    """

    def deserialize(self, value: str, **_kwargs: Any) -> bytes:
        """Deserialize a bytes parameter."""
        # Extract all integer tokens from the string
        tokens = [int(x) for x in re.findall(r"-?\d+", value)]

        # Convert each token to a signed 32-bit integer and concatenate them into bytes
        return b"".join(struct.pack("i", token) for token in tokens)

    def serialize(self, value: bytes, **_kwargs: Any) -> str:
        """Serialize a bytes parameter."""
        # Pad the data to a multiple of 4 bytes
        value += b"\x00" * (-len(value) % 4)

        # Convert each signed 32-bit integer in the byte array to a string token
        tokens = [
            str(struct.unpack("i", value[i : i + 4])[0])
            for i in range(0, len(value), 4)
        ]

        # Join the tokens with commas and wrap in curly braces
        return "{" + ",".join(tokens) + "}"


class DateTimeConverter(Converter):
    """A datetime converter.

    Vantage "Time" parameters are encoded as a Unix timestamp.
    """

    def deserialize(self, value: str, **_kwargs: Any) -> dt.datetime:
        """Deserialize a datetime value."""
        return dt.datetime.fromtimestamp(int(value), dt.timezone.utc)

    def serialize(self, value: dt.datetime, **_kwargs: Any) -> str:
        """Serialize a datetime value."""
        return str(int(value.timestamp()))


class IntEnumConverter(Converter):
    """An IntEnum converter.

    Vantage "Enum" values are encoded as either their integer or string
    representation.

    String representations are either lowercase single words or PascalCase.
    """

    def deserialize(self, value: str, **kwargs: Any) -> Any:
        """Deserialize an IntEnum value."""
        enum_type = kwargs.get("data_type")
        if enum_type is None or not issubclass(enum_type, IntEnum):
            raise ValueError("IntEnumConverter requires a data_type argument")

        # Handle integer representations of enum values
        if value.isdigit():
            return enum_type(int(value))

        # Handle string representations of enum values.
        return enum_type[value]

    def serialize(self, value: Any, **kwargs: Any) -> str:
        """Serialize an IntEnum value."""
        # The API accepts either the integer or string representation of the enum
        # We'll always serialize to the integer representation for simplicity
        return str(value.value)


# Create a global converter registry, and register the default converters
converter = ConverterRegistry()
converter.register(str, StringConverter())
converter.register(int, IntConverter())
converter.register(bool, BoolConverter())
converter.register(float, FloatConverter())
converter.register(bytes, BytesConverter())
converter.register(Decimal, DecimalConverter())
converter.register(dt.datetime, DateTimeConverter())
converter.register(IntEnum, IntEnumConverter())



"""IConfiguration.GetObject method definition."""

from dataclasses import dataclass, field

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetObject:
    """IConfiguration.GetObject method definition."""

    interface = "IConfiguration"

    call: list[int] | None = field(
        default=None,
        metadata={
            "wrapper": "call",
            "name": "VID",
            "type": "Element",
        },
    )

    result: list[ObjectChoice] | None = field(
        default=None,
        metadata={
            "wrapper": "return",
            "name": "Object",
            "type": "Element",
        },
    )



"""DryContact object."""

from dataclasses import dataclass, field

from .location_object import LocationObject
from .types import Parent


@dataclass(kw_only=True)
class DryContact(LocationObject):
    """DryContact object."""

    parent: Parent
    down: int = 0
    up: int = 0
    reverse_polarity: bool
    hold_on_time: float = 0

    # State
    triggered: bool | None = field(default=None, metadata={"type": "Ignore"})



"""ILogin.Login method definition."""

from dataclasses import dataclass, field


@dataclass
class Login:
    """ILogin.Login method definition."""

    interface = "ILogin"

    @dataclass
    class Params:
        """Method parameters."""

        user: str
        password: str

    call: Params | None = field(
        default=None,
        metadata={
            "name": "call",
        },
    )

    result: bool | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""ObjectChoice type definition."""

from dataclasses import dataclass, field
from functools import cache
from importlib import import_module
from inspect import getmembers, isclass

from aiovantage.objects import SystemObject


@cache
def get_all_object_choices() -> list[dict[str, str | type[SystemObject]]]:
    """Get all object choices from a module."""
    # Load all SystemObject types into memory
    module = import_module("aiovantage.objects")

    # Build and return choices list
    return [
        {"name": cls.vantage_type(), "type": cls}
        for _, cls in getmembers(module, isclass)
        if issubclass(cls, SystemObject)
    ]


@dataclass
class ObjectChoice:
    """Wildcard type that can be used to represent any object."""

    id: int = field(
        metadata={
            "name": "VID",
            "type": "Attribute",
        }
    )

    choice: object = field(
        metadata={
            "type": "Wildcard",
            "choices": get_all_object_choices(),
        },
    )



"""IConfiguration.GetFilterResults method definition."""

from dataclasses import dataclass, field

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetFilterResults:
    """IConfiguration.GetFilterResults method definition."""

    interface = "IConfiguration"

    @dataclass
    class Params:
        """Method parameters."""

        h_filter: int = field(metadata={"name": "hFilter"})
        count: int = 50
        whole_object: bool = True

    call: Params | None = field(
        default=None,
        metadata={
            "name": "call",
        },
    )

    result: list[ObjectChoice] | None = field(
        default=None,
        metadata={
            "wrapper": "return",
            "name": "Object",
            "type": "Element",
        },
    )



"""Interface for querying and controlling blinds."""

from dataclasses import dataclass
from decimal import Decimal

from .base import Interface


class BlindInterface(Interface):
    """Interface for querying and controlling blinds."""

    @dataclass
    class BlindState:
        """The state of a blind."""

        is_moving: bool
        start_pos: Decimal
        end_pos: Decimal
        transition_time: Decimal
        start_time: int

    @dataclass
    class TravelTimes:
        """The travel times of a blind."""

        rcode: int
        open_time: Decimal
        close_time: Decimal

    method_signatures = {
        "Blind.GetPosition": Decimal,
        "Blind.GetPositionHW": Decimal,
        "Blind.GetTiltAngle": int,
        "Blind.GetTiltAngleHW": int,
        "Blind.IsTiltAvailable": bool,
        "Blind.GetBlindState": BlindState,
        "Blind.GetUpperLimit": Decimal,
        "Blind.GetUpperLimitHW": Decimal,
        "Blind.GetLowerLimit": Decimal,
        "Blind.GetLowerLimitHW": Decimal,
        "Blind.GetTravelTimes": TravelTimes,
    }

    async def open(self, vid: int) -> None:
        """Open a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Open
        # -> R:INVOKE <id> <rcode> Blind.Open
        await self.invoke(vid, "Blind.Open")

    async def close(self, vid: int) -> None:
        """Close a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Close
        # -> R:INVOKE <id> <rcode> Blind.Close
        await self.invoke(vid, "Blind.Close")

    async def stop(self, vid: int) -> None:
        """Stop a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Stop
        # -> R:INVOKE <id> <rcode> Blind.Stop
        await self.invoke(vid, "Blind.Stop")

    async def set_position(
        self, vid: int, position: float, *, sw: bool = False
    ) -> None:
        """Set the position of a blind.

        Args:
            vid: The Vantage ID of the blind.
            position: The position to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetPosition <position>
        # -> R:INVOKE <id> <rcode> Blind.SetPosition <position>
        await self.invoke(
            vid, "Blind.SetPositionSW" if sw else "Blind.SetPosition", position
        )

    async def get_position(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the position of a blind.

        Args:
            vid: The Vantage ID of the blind.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The position of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetPosition
        # -> R:INVOKE <id> <position (0-100.000)> Blind.GetPosition
        return await self.invoke(
            vid, "Blind.GetPositionHW" if hw else "Blind.GetPosition"
        )

    async def set_tilt_angle(self, vid: int, angle: int, *, sw: bool = False) -> None:
        """Set the tilt angle of a blind.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle to set the blind to, from -100 to 100.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetTiltAngle <angle>
        # -> R:INVOKE <id> <rcode> Blind.SetTiltAngle <angle>
        await self.invoke(
            vid, "Blind.SetTiltAngleSW" if sw else "Blind.SetTiltAngle", angle
        )

    async def get_tilt_angle(self, vid: int, *, hw: bool = False) -> int:
        """Get the tilt angle of a blind.

        Args:
            vid: The Vantage ID of the blind.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The tilt angle of the blind, from -100 to 100.
        """
        # INVOKE <id> Blind.GetTiltAngle
        # -> R:INVOKE <id> <angle (-100-100)> Blind.GetTiltAngle
        return await self.invoke(
            vid, "Blind.GetTiltAngleHW" if hw else "Blind.GetTiltAngle"
        )

    async def tilt_clockwise(self, vid: int, angle: int) -> None:
        """Tilt the blinds clockwise by the specified angle.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltClockwise <angle>
        await self.invoke(vid, "Blind.TiltClockwise", angle)

    async def tilt_counter_clockwise(self, vid: int, angle: int) -> None:
        """Tilt the shades counter-clockwise by the specified angle.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltCounterClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltCounterClockwise <angle>
        await self.invoke(vid, "Blind.TiltCounterClockwise", angle)

    async def is_tilt_available(self, vid: int) -> bool:
        """Check if the blind can tilt in its current state.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            Whether the blind supports tilting.
        """
        # INVOKE <id> Blind.IsTiltAvailable
        # -> R:INVOKE <id> <available (0/1)> Blind.IsTiltAvailable
        return await self.invoke(vid, "Blind.IsTiltAvailable")

    async def set_tilt_available(self, vid: int, available: bool) -> None:
        """Set the cached tilt availability of a blind.

        Args:
            vid: The Vantage ID of the blind.
            available: Whether the blind supports tilting.
        """
        # INVOKE <id> Blind.SetTiltAvailableSW <available>
        # -> R:INVOKE <id> <rcode> Blind.SetTiltAvailableSW <available>
        await self.invoke(vid, "Blind.SetTiltAvailableSW", available)

    async def get_blind_state(self, vid: int) -> BlindState:
        """Get the state of a blind.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The state of the blind.
        """
        # INVOKE <id> Blind.GetBlindState
        # -> R:INVOKE <id> <moving> Blind.GetBlindState <start> <end> <transitionTime> <startTime>
        return await self.invoke(vid, "Blind.GetBlindState")

    async def set_upper_limit(
        self, vid: int, limit: Decimal, *, sw: bool = False
    ) -> None:
        """Set the upper limit of a blind.

        Args:
            vid: The Vantage ID of the blind.
            limit: The upper limit to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetUpperLimit <limit>
        # -> R:INVOKE <id> <rcode> Blind.SetUpperLimit <limit>
        await self.invoke(
            vid, "Blind.SetUpperLimitSW" if sw else "Blind.SetUpperLimit", limit
        )

    async def get_upper_limit(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the upper limit of a blind.

        Args:
            vid: The Vantage ID of the blind.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The upper limit of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetUpperLimit
        # -> R:INVOKE <id> <limit (0-100.000)> Blind.GetUpperLimit
        return await self.invoke(
            vid, "Blind.GetUpperLimitHW" if hw else "Blind.GetUpperLimit"
        )

    async def set_lower_limit(
        self, vid: int, limit: Decimal, *, sw: bool = False
    ) -> None:
        """Set the lower limit of a blind.

        Args:
            vid: The Vantage ID of the blind.
            limit: The lower limit to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetLowerLimit <limit>
        # -> R:INVOKE <id> <rcode> Blind.SetLowerLimit <limit>
        await self.invoke(
            vid, "Blind.SetLowerLimitSW" if sw else "Blind.SetLowerLimit", limit
        )

    async def get_lower_limit(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the lower limit of a blind.

        Args:
            vid: The Vantage ID of the blind.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The lower limit of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetLowerLimit
        # -> R:INVOKE <id> <limit (0-100.000)> Blind.GetLowerLimit
        return await self.invoke(
            vid, "Blind.GetLowerLimitHW" if hw else "Blind.GetLowerLimit"
        )

    async def get_travel_times(self, vid: int) -> TravelTimes:
        """Get the travel times of a blind.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The travel times of the blind.
        """
        # INVOKE <id> Blind.GetTravelTimes
        # -> R:INVOKE <id> <openTime> <closeTime> Blind.GetTravelTimes
        return await self.invoke(vid, "Blind.GetTravelTimes")



"""Somfy URTSI 2 port device."""

from dataclasses import dataclass

from .port_device import PortDevice


@dataclass(kw_only=True)
class SomfyURTSI2Port(PortDevice):
    """Somfy URTSI 2 port device."""

    class Meta:
        name = "Somfy.URTSI_2_PORT"



"""LoadGroup object."""

from dataclasses import dataclass, field

from .location_object import LocationObject


@dataclass(kw_only=True)
class LoadGroup(LocationObject):
    """LoadGroup object."""

    load_table: list[int] = field(
        default_factory=list, metadata={"name": "Load", "wrapper": "LoadTable"}
    )

    # State
    level: float | None = field(default=None, metadata={"type": "Ignore"})

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



"""RGB load object."""

from dataclasses import dataclass, field
from enum import Enum

from .location_object import LocationObject
from .types import Parent


@dataclass(kw_only=True)
class RGBLoadBase(LocationObject):
    """RGB load base class."""

    class ColorType(Enum):
        RGB = "RGB"
        RGBW = "RGBW"
        HSL = "HSL"
        HSIC = "HSIC"
        CCT = "CCT"
        COLOR_CHANNEL = "Color Channel"

    parent: Parent
    color_type: ColorType
    min_temp: int
    max_temp: int

    # State
    hsl: tuple[int, int, int] | None = field(default=None, metadata={"type": "Ignore"})
    rgb: tuple[int, int, int] | None = field(default=None, metadata={"type": "Ignore"})
    rgbw: tuple[int, int, int, int] | None = field(
        default=None, metadata={"type": "Ignore"}
    )
    level: int | None = field(default=None, metadata={"type": "Ignore"})
    color_temp: int | None = field(default=None, metadata={"type": "Ignore"})

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)

    @property
    def is_rgb(self) -> bool:
        """Return True if the load is an RGB(W) load."""
        return self.color_type in (
            self.ColorType.RGB,
            self.ColorType.RGBW,
            self.ColorType.HSL,
        )

    @property
    def is_cct(self) -> bool:
        """Return True if the load is a CCT load."""
        return self.color_type == self.ColorType.CCT



"""Interface for querying and controlling sensors."""

from decimal import Decimal

from .base import Interface


class SensorInterface(Interface):
    """Interface for querying and controlling sensors."""

    method_signatures = {
        "Sensor.GetLevel": Decimal,
        "Sensor.GetLevelHW": Decimal,
        "Sensor.GetHighRange": Decimal,
        "Sensor.GetLowRange": Decimal,
        "Sensor.GetHoldOnTime": Decimal,
        "Sensor.IsTracking": bool,
        "Sensor.GetTrackingDelta": Decimal,
        "Sensor.GetTrackingMin": Decimal,
        "Sensor.GetTrackingMax": Decimal,
    }

    async def get_level(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the level of a sensor.

        Args:
            vid: The Vantage ID of the sensor.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the sensor.
        """
        # INVOKE <id> Sensor.GetLevel
        # -> R:INVOKE <id> <level (0-100)> Sensor.GetLevel
        return await self.invoke(vid, "Sensor.GetLevelHW" if hw else "Sensor.GetLevel")

    async def set_level(self, vid: int, level: Decimal) -> None:
        """Set the level of a sensor.

        Args:
            vid: The Vantage ID of the sensor.
            level: The level to set the sensor to.
        """
        # INVOKE <id> Sensor.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Sensor.SetLevel <level (0-100)>
        await self.invoke(vid, "Sensor.SetLevel", level)

    async def get_high_range(self, vid: int) -> Decimal:
        """Get the high range of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The high range of the sensor.
        """
        # INVOKE <id> Sensor.GetHighRange
        # -> R:INVOKE <id> <high range> Sensor.GetHighRange
        return await self.invoke(vid, "Sensor.GetHighRange")

    async def get_low_range(self, vid: int) -> Decimal:
        """Get the low range of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The low range of the sensor.
        """
        # INVOKE <id> Sensor.GetLowRange
        # -> R:INVOKE <id> <low range> Sensor.GetLowRange
        return await self.invoke(vid, "Sensor.GetLowRange")

    async def get_hold_on_time(self, vid: int) -> Decimal:
        """Get the hold on time of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The hold on time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetHoldOnTime
        # -> R:INVOKE <id> <hold on time> Sensor.GetHoldOnTime
        return await self.invoke(vid, "Sensor.GetHoldOnTime")

    async def is_tracking(self, vid: int) -> bool:
        """Get whether the sensor is tracking.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            Whether the sensor is tracking.
        """
        # INVOKE <id> Sensor.IsTracking
        # -> R:INVOKE <id> <tracking (0/1)> Sensor.IsTracking
        return await self.invoke(vid, "Sensor.IsTracking")

    async def get_tracking_delta(self, vid: int) -> Decimal:
        """Get the tracking delta of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The tracking delta of the sensor.
        """
        # INVOKE <id> Sensor.GetTrackingDelta
        # -> R:INVOKE <id> <tracking delta> Sensor.GetTrackingDelta
        return await self.invoke(vid, "Sensor.GetTrackingDelta")

    async def get_tracking_min(self, vid: int) -> Decimal:
        """Get the tracking min time of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The tracking min time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetTrackingMin
        # -> R:INVOKE <id> <tracking min> Sensor.GetTrackingMin
        return await self.invoke(vid, "Sensor.GetTrackingMin")

    async def get_tracking_max(self, vid: int) -> Decimal:
        """Get the tracking max time of a sensor.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The tracking max time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetTrackingMax
        # -> R:INVOKE <id> <tracking max> Sensor.GetTrackingMax
        return await self.invoke(vid, "Sensor.GetTrackingMax")



"""Task object."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass(kw_only=True)
class Task(SystemObject):
    """Task object."""

    # State
    is_running: bool | None = field(default=None, metadata={"type": "Ignore"})
    state: int | None = field(default=None, metadata={"type": "Ignore"})



"""Interface for querying and controlling anemo (wind) sensors."""

from decimal import Decimal

from .base import Interface


class AnemoSensorInterface(Interface):
    """Interface for querying and controlling anemo (wind) sensors."""

    method_signatures = {
        "AnemoSensor.GetSpeed": Decimal,
        "AnemoSensor.GetSpeedHW": Decimal,
    }

    async def get_speed(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the speed of an anemo sensor.

        Args:
            vid: The ID of the sensor.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The speed of the anemo sensor, in mph.
        """
        # INVOKE <id> AnemoSensor.GetSpeed
        # -> R:INVOKE <id> <speed> AnemoSensor.GetSpeed
        return await self.invoke(
            vid, "AnemoSensor.GetSpeedHW" if hw else "AnemoSensor.GetSpeed"
        )

    async def set_speed(self, vid: int, speed: Decimal, *, sw: bool = False) -> None:
        """Set the speed of an anemo sensor.

        Args:
            vid: The ID of the sensor.
            speed: The speed to set, in mph.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> AnemoSensor.SetSpeed <speed>
        # -> R:INVOKE <id> <rcode> AnemoSensor.SetSpeed <speed>
        await self.invoke(
            vid, "AnemoSensor.SetSpeedSW" if sw else "AnemoSensor.SetSpeed", speed
        )



"""Light sensor object."""

from dataclasses import dataclass, field
from decimal import Decimal

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class LightSensor(Sensor):
    """Light sensor object."""

    parent: Parent

    # State
    level: Decimal | None = field(default=None, metadata={"type": "Ignore"})



"""Interface for querying and controlling sensors."""

from decimal import Decimal

from .base import Interface


class TemperatureInterface(Interface):
    """Interface for querying and controlling sensors."""

    method_signatures = {
        "Temperature.GetValue": Decimal,
        "Temperature.GetValueHW": Decimal,
    }

    async def get_value(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the value of a temperature sensor.

        Args:
            vid: The Vantage ID of the temperature object.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the temperature sensor, in degrees Celsius.
        """
        # INVOKE <id> Temperature.GetValue
        # -> R:INVOKE <id> <temp> Temperature.GetValue
        return await self.invoke(
            vid, "Temperature.GetValueHW" if hw else "Temperature.GetValue"
        )

    async def set_value(self, vid: int, value: Decimal, *, sw: bool = False) -> None:
        """Set the value of a temperature sensor.

        Args:
            vid: The Vantage ID of the temperature object.
            value: The value to set the sensor to.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Temperature.SetValue <value>
        # -> R:INVOKE <id> <rcode> Temperature.SetValue <value>
        await self.invoke(
            vid, "Temperature.SetValueSW" if sw else "Temperature.SetValue", value
        )



"""Interface for querying and controlling light sensors."""

from decimal import Decimal

from .base import Interface


class LightSensorInterface(Interface):
    """Interface for querying and controlling light sensors."""

    method_signatures = {
        "LightSensor.GetLevel": Decimal,
        "LightSensor.GetLevelHW": Decimal,
    }

    async def get_level(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the level of a light sensor.

        Args:
            vid: The Vantage ID of the light sensor.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the light sensor, in foot-candles.
        """
        # INVOKE <id> LightSensor.GetLevel
        # -> R:INVOKE <id> <level> LightSensor.GetLevel
        return await self.invoke(
            vid, "LightSensor.GetLevelHW" if hw else "LightSensor.GetLevel"
        )

    async def set_level(self, vid: int, level: Decimal) -> None:
        """Set the level of a light sensor.

        Args:
            vid: The Vantage ID of the light sensor.
            level: The level to set, in foot-candles.
        """
        # INVOKE <id> LightSensor.SetLevel <level>
        # -> R:INVOKE <id> <rcode> LightSensor.SetLevel <level>
        await self.invoke(vid, "LightSensor.SetLevel", level)



"""Interface for controller introspection."""

from dataclasses import dataclass
from enum import IntEnum

from .base import Interface


class IntrospectionInterface(Interface):
    """Interface for controller introspection."""

    class Firmware(IntEnum):
        """Firmware image."""

        Kernel = 0
        RootFs = 1
        Application = 2

    class LicenseType(IntEnum):
        """License type."""

        Unknown = 0
        Equinox = 1
        All = -1

    @dataclass
    class FirmwareVersion:
        """A firmware version response."""

        rcode: int
        image: "IntrospectionInterface.Firmware"
        version: str
        size: int

    @dataclass
    class LicenseInfo:
        """A license info response."""

        rcode: int
        type: "IntrospectionInterface.LicenseType"
        used: int
        total: int

    method_signatures = {
        "Introspection.GetAppControllers": str,
        "Introspection.GetFirmwareVersion": FirmwareVersion,
        "Introspection.GetLicenseInfo": LicenseInfo,
    }

    async def get_app_controllers(self, vid: int) -> str:
        """Get a list of controllers in application mode, exclidng this controller.

        Args:
            vid: The Vantage ID of the controller.

        Returns:
            A comma-separated list of controller numbers.
        """
        # INVOKE <id> Introspection.GetAppControllers
        # -> R:INVOKE <id> <rcode> Introspection.GetAppControllers <controllers>
        return await self.invoke(vid, "Introspection.GetAppControllers")

    async def get_firmware_version(self, vid: int, image: Firmware) -> FirmwareVersion:
        """Get the firmware version.

        Args:
            vid: The Vantage ID of the controller.
            image: The firmware image to get the version of.
        """
        # INVOKE <id> Introspection.GetFirmwareVersion <image>
        # -> R:INVOKE <id> <rcode> Introspection.GetFirmwareVersion <image> <version>
        return await self.invoke(vid, "Introspection.GetFirmwareVersion", image)

    async def get_license_info(self, vid: int, type: LicenseType) -> LicenseInfo:
        """Get license information.

        Args:
            vid: The Vantage ID of the controller.
            type: The license type to get information for.
        """
        # INVOKE <id> Introspection.GetLicenseInfo <type>
        # -> R:INVOKE <id> <rcode> Introspection.GetLicenseInfo <type> <used> <total>
        return await self.invoke(vid, "Introspection.GetLicenseInfo", type)

    # Convenience functions, not part of the interface
    async def get_application_version(self, vid: int) -> str:
        """Get the application firmware version."""
        response = await self.get_firmware_version(vid, self.Firmware.Application)

        return response.version



"""Keypad Station."""

from dataclasses import dataclass

from .station_object import StationObject
from .types import Parent


@dataclass(kw_only=True)
class Keypad(StationObject):
    """Keypad Station."""

    parent: Parent



"""Base class for command client interfaces."""

from typing import Any, TypeVar, get_type_hints, overload

from aiovantage.command_client import CommandClient
from aiovantage.command_client.types import converter

T = TypeVar("T")


class Interface:
    """Base class for command client object interfaces."""

    method_signatures: dict[str, type[Any]] = {}

    def __init__(self, client: CommandClient) -> None:
        """Initialize an object interface for standalone use.

        Args:
            client: The command client to use.
        """
        self._command_client = client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client."""
        return self._command_client

    @overload
    async def invoke(self, vid: int, method: str, *params: Any) -> Any: ...

    @overload
    async def invoke(
        self, vid: int, method: str, *params: Any, as_type: type[T]
    ) -> T: ...

    async def invoke(
        self, vid: int, method: str, *params: Any, as_type: type[T] | None = None
    ) -> T | Any:
        """Invoke a method on an object, and return the parsed response.

        Args:
            vid: The VID of the object to invoke the command on.
            method: The method to invoke.
            params: The parameters to send with the method.
            as_type: The type to cast the response to.

        Returns:
            A parsed response, or None if no response was expected.
        """
        # Send the command
        response = await self.command_client.command(
            "INVOKE", vid, method, *params, force_quotes=True
        )

        # Break the response into tokens
        _id, result, _method, *args = response.args

        # Parse the response
        signature = as_type or self.get_method_signature(method)
        return _parse_object_response(result, *args, as_type=signature)

    @classmethod
    def get_method_signature(cls, method: str) -> type[Any] | None:
        """Get the signature of a method."""
        for klass in cls.__mro__:
            if issubclass(klass, Interface) and method in klass.method_signatures:
                return klass.method_signatures[method]

    @classmethod
    def parse_object_status(cls, method: str, result: str, *args: str) -> Any:
        """Handle an object interface status message.

        Args:
            method: The method that was invoked.
            result: The result of the command.
            args: The arguments that were sent with the command.
        """
        # Parse the response
        signature = cls.get_method_signature(method)
        return _parse_object_response(result, *args, as_type=signature)


def _parse_object_response(
    result: str, *args: str, as_type: type[T] | None
) -> T | None:
    """Parse an object interface response message.

    Args:
        result: The "result" of the command, aka the return value.
        args: The arguments that were sent with the command, which may have been modified.
        as_type: The expected return type of the method.

    Returns:
        A response parsed into the expected type.
    """
    # -> R:INVOKE <id> <result> <Interface.Method> <arg1> <arg2> ...
    # -> EL: <id> <Interface.Method> <result> <arg1> <arg2> ...
    # -> S:STATUS <id> <Interface.Method> <result> <arg1> <arg2> ...

    # Return early if the expected type is NoneType
    if as_type in (None, type(None)):
        return None

    # Otherwise, parse the result into the expected type
    if type_hints := get_type_hints(as_type):
        # Some methods return multiple values, in the "return" field and in the
        # "params" fields. This adds support for parsing multiple values into
        # a dataclass or NamedTuple.
        parsed_values: list[Any] = []
        for arg, klass in zip([result, *args], type_hints.values(), strict=True):
            parsed_value = converter.deserialize(klass, arg)
            parsed_values.append(parsed_value)

        parsed_response = as_type(*parsed_values)
    else:
        # Simple string responses contain the string value in the first argument
        param = args[0] if as_type is str else result
        parsed_response = converter.deserialize(as_type, param)

    # Return the parsed result
    return parsed_response



"""Interface for querying and controlling power sensors."""

from decimal import Decimal

from .base import Interface


class PowerSensorInterface(Interface):
    """Interface for querying and controlling power sensors."""

    method_signatures = {
        "CurrentSensor.GetPower": Decimal,
        "CurrentSensor.GetPowerHW": Decimal,
    }

    async def get_power(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the value of a power sensor.

        Args:
            vid: The Vantage ID of the power sensor.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the power sensor, in Watts.
        """
        # INVOKE <id> PowerSensor.GetPower
        # -> R:INVOKE <id> <level> PowerSensor.GetPower
        return await self.invoke(
            vid, "PowerSensor.GetPowerHW" if hw else "PowerSensor.GetPower"
        )

    async def set_power(self, vid: int, value: Decimal, *, sw: bool = False) -> None:
        """Set the value of a power sensor.

        Args:
            vid: The Vantage ID of the power sensor.
            value: The value to set, in Watts.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> PowerSensor.SetPower <level>
        # -> R:INVOKE <id> <rcode> PowerSensor.SetPower <level>
        await self.invoke(
            vid, "PowerSensor.SetPowerSW" if sw else "PowerSensor.SetPower", value
        )



"""Interface for querying and controlling current sensors."""

from decimal import Decimal

from .base import Interface


class CurrentSensorInterface(Interface):
    """Interface for querying and controlling current sensors."""

    method_signatures = {
        "CurrentSensor.GetCurrent": Decimal,
        "CurrentSensor.GetCurrentHW": Decimal,
    }

    async def get_current(self, vid: int, *, hw: bool = False) -> Decimal:
        """Get the value of a current sensor.

        Args:
            vid: The Vantage ID of the current sensor.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the current sensor, in Amps.
        """
        # INVOKE <id> CurrentSensor.GetCurrent
        # -> R:INVOKE <id> <level> CurrentSensor.GetCurrent
        return await self.invoke(
            vid, "CurrentSensor.GetCurrentHW" if hw else "CurrentSensor.GetCurrent"
        )

    async def set_current(self, vid: int, value: Decimal, *, sw: bool = False) -> None:
        """Set the value of a current sensor.

        Args:
            vid: The Vantage ID of the current sensor.
            value: The value to set, in Amps.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> CurrentSensor.SetCurrent <level>
        # -> R:INVOKE <id> <rcode> CurrentSensor.SetCurrent <level>
        await self.invoke(
            vid,
            "CurrentSensor.SetCurrentSW" if sw else "CurrentSensor.SetCurrent",
            value,
        )
